#include "mazeUtil.h"
#include <vector>

using std::vector;
using namespace std;

//

struct comparer 
{
	bool operator()(const pos &a, const pos &b) const 
	{
		return a.step < b.step;
	}
} _comparer;

//

mazeUtil::mazeUtil(void)
{
}

mazeUtil::~mazeUtil(void)
{
}

void mazeUtil::generate(maze* $pMaze)
{
	clock_t startTime = clock();

	// setup...

	_m = $pMaze;
	_m->fill(true);

	_w = _m->getWidth() - 1;
	_h = _m->getHeight() - 1;

	_scratch = vector<int>();
	_scratch.push_back(0);
	_scratch.push_back(1);
	_scratch.push_back(2);
	_scratch.push_back(3);

	// init cells

	// xxx don't know how to re-instantiate this 2-dimensional array
	// _visited = new bool[100][100];


	for (int y = 0; y < _h; y++) {
	  for (int x = 0; x < _w; x++) {
		_visited[x][y] = false;
	  }
	}

	_crumbs = vector<pos>();
	_history = vector<pos>();

	//

	pos p = doStartPosAndBreakWall();

	_visited[p.x][p.y] = true;
	_crumbs.push_back(p.clone());
	_history.push_back(p.clone());

	do
	{

		pos pos = _crumbs[_crumbs.size()-1]; /// may be wrong
		_crumbs.pop_back();

		///_crumbs.erase(_crumbs.size()-1);
		// println("removed crumb " + pos.x + "," + pos.y);
		// printPosArrayList(_crumbs);

		makePath(pos);
	}
	while (_crumbs.size() > 0);

	doEndPosAndBreakWall();

	//

	clock_t endTime = clock();
	cout << "Elapsed:" << double(endTime - startTime) / CLOCKS_PER_SEC << " seconds" << endl;
}

void mazeUtil::makePath(pos p)
{
	// println("makePath " + p.x + " " + p.y);

	while(true)
	{
	  int dir = getLegalDirection(p);
	  
	  if (dir == DIR_NONE) 
	  {
		return;
	  }

	  switch(dir)
	  {
		case DIR_LEFT:
		 _m->setWallV(p.x, p.y, false); 
		  p.x--; 
		  break;
		case DIR_RIGHT: 
		 _m->setWallV(p.x+1, p.y, false); 
		  p.x++; 
		  break;
		case DIR_UP:
		 _m->setWallH(p.y, p.x, false);
		  p.y--;
		  break;
		case DIR_DOWN:
		 _m->setWallH(p.y+1, p.x, false);
		  p.y++; 
		  break;
	  }
	  
	  _visited[p.x][p.y] = true;

	  p.step = _crumbs.size()+1;
	  
	  _crumbs.push_back(p.clone());
	  _history.push_back(p.clone());
	  
	  // println("moved. pos is now " + p.x + "," + p.y + " step: " + p.step);
	  // printposArrayList(_crumbs);
	}
}

void mazeUtil::printPosArrayList(vector<pos> a)
{
	stringstream s;
	for (int i = 0; i < a.size(); i++)
	{
		//printf("element %i: %i,%i,%i\n", i, a[i].x, a[i].y, a[i].step);
		s << " | " << i << ": " << a[i].x << "," << a[i].y << "," << a[i].step;
	}

	cout << s.str() << endl;
}


int mazeUtil::getLegalDirection(pos p)
{
	int *dirs = getRandomDirectionsList();

	for (int i = 0; i < 4; i++)
	{
	  int dir = dirs[i];
	  
	  switch(dir)
	  {
		case DIR_LEFT:
		  if (p.x > 0 && ! _visited[p.x-1][p.y]) return dir;
		  break;
		case DIR_RIGHT:
		  if (p.x < _w-1 && ! _visited[p.x+1][p.y]) return dir;
		  break;
		case DIR_UP:
		  if (p.y > 0 && ! _visited[p.x][p.y-1]) return dir;
		  break;
		case DIR_DOWN:
		  if (p.y < _h-1 && ! _visited[p.x][p.y+1]) return dir;
		  break;
	  }
	}

	return DIR_NONE;
}

int* mazeUtil::getRandomDirectionsList()
{
	random_shuffle(_scratch.begin(), _scratch.end());

	int* a = new int[4];
	for (int i = 0; i < 4; i++)
	{
	  a[i] = _scratch[i]; /// may be wrong
	}
	return a;
}


pos mazeUtil::doStartPosAndBreakWall()
{
	int dx, dy;

	int edge = rand() % 4;

	switch (edge)
	{
	  case DIR_LEFT:
		dx = 0;
		dy = rand() % (_m->getHeight() - 1);
		_m->setWallV(dx, dy, false);
		_m->setStartDirection(DIR_RIGHT);
		break;

	  case DIR_RIGHT:
		dx = _m->getWidth() - 2;
		dy = rand() % (_m->getHeight() - 1);
		_m->setWallV(dx+1, dy, false);
		_m->setStartDirection(DIR_LEFT);
		break;

	  case DIR_UP:
		dy = 0;
		dx = rand() % (_m->getWidth() - 1);
		_m->setWallH(dy, dx, false);
		_m->setStartDirection(DIR_DOWN);
		break;

	  case DIR_DOWN:
		dy = _m->getHeight() - 2;
		dx = rand() % (_m->getWidth() - 1);
		_m->setWallH(dy+1, dx, false);
		_m->setStartDirection(DIR_UP);
		break;
	}

	_m->setStartPos( pos(dx,dy,0) );

	return pos(dx,dy,0);
}


void mazeUtil::doEndPosAndBreakWall()
{
	sort(_history.begin(), _history.end(), _comparer); // wow!

	pos p;

	while (true)
	{
		p = _history[ _history.size()-1 ];
		_history.pop_back(); /// this may be wrong

		if (p.x == 0) // like dir_left
		{
			_m->setWallV(p.x, p.y, false);
			break;
		} 
		else if (p.x == _m->getWidth()-2) // like dir_right
		{      
			_m->setWallV(p.x+1, p.y, false);
			break;
		}
		else if (p.y == 0) // like dir_up
		{
			_m->setWallH(p.y, p.x, false);
			break;
		}
		else if (p.y == _m->getHeight()-2) // like dir_down
		{
			_m->setWallH(p.y+1, p.x, false);
			break;
		}
		// else keep going
	}

	_m->setEndPos(p);
}

bool mazeUtil::isFull()
{
	for (int y = 0; y < _h; y++)
	{
	  for (int x = 0; x < _w; x++)
	  {
		if (! _visited[x][y]) return false;
	  }
	}
	return true;
}

bool mazeUtil::compareAscending(pos $a, pos $b) 
{ 
	return ($a.step > $b.step); 
}
